import { DynamoDBClient, AttributeValue } from '@aws-sdk/client-dynamodb';
import { VError, Options } from 'verror';

/**
 * The structure of your primary index. Your table
 * must have a partition and sort keys, both of
 * type 'S' (string)
 */
declare type PrimaryIndexLayout = {
    /**
     * The name of the primary partition key
     */
    partitionKey: string;
    /**
     * The name of the primary sort key
     */
    sortKey: string;
};
/**
 * The structure of a Global Secondary Index
 */
declare type SecondaryIndexLayout = {
    /**
     * The name of the secondary index
     */
    indexName: string;
    /**
     * The name of the secondary index partition key
     */
    partitionKey: string;
    /**
     * The name of the secondary index sort key. If
     * your index does not have one, leave this as
     * `undefined`
     */
    sortKey?: string;
};
/**
 * Defines the structure of the underlying table of a collection,
 * such as its table name, primary keys, sort keys and TTL
 * attribute name.
 *
 * You can also customise the index name separator
 */
interface CollectionLayout {
    /** Name of the table */
    tableName: string;
    /** Layout of the primary key */
    primaryKey: PrimaryIndexLayout;
    /**
     * Optional secondary find keys for additional lookups
     */
    findKeys?: SecondaryIndexLayout[];
    /**
     * The name of the attribute configured as the
     * TimeToLiveSpecification attribute
     */
    ttlAttribute?: string;
    /**
     * The separator to use in index keys (defaults to `|-|`)
     */
    indexKeySeparator?: string;
}

/**
 * A path to a key in a document. Each path element should be
 * a string. For example, a field called 'users' at the top level is
 * simply `['userId']`, while a nested field at 'description.title[0]'
 * is `['description', 'title', '0']`
 */
declare type KeyPath = string[];
/**
 * A function used to normalise a key value before it is packed into
 * an index value.
 *
 * This is useful when you want to index your fields differently to how
 * they're stored e.g. lowercase values before they are put in the
 * index.
 */
declare type NormaliserFunction = (path: KeyPath, value: string) => string;
/**
 * Options for an access pattern.
 */
declare type AccessPatternOptions = {
    /** A normaliser for values stored in the access pattern's index */
    stringNormalizer?: NormaliserFunction;
};
/**
 * A **access pattern** defines how to copy values from the entities
 * in a collection into an index. They are used to permit indexed
 * queryies with the [[find]] operation.
 *
 * In order to index values, you must first have defined a GSI in
 * your [[CollectionLayout|collection's layout]].
 */
interface AccessPattern {
    /** The name of the index in your [[CollectionLayout|collection's layout]] */
    indexName: string;
    /** The [[KeyPath|key paths]] to extract and store in the partition key.
     *
     * Key paths specified for a partition key must be present in documents
     * stored in the associated [[Collection]].
     */
    partitionKeys: KeyPath[];
    /** The [[KeyPath|key paths]] to extract and store in the sort key.
     *
     * Leave this value unspecified if your GSI does not have a sort key. If it does
     * but you don't want a sort key, populate it with an empty array
     */
    sortKeys?: KeyPath[];
    /** Options for the index, such as string normalisers */
    options?: AccessPatternOptions;
}

interface CommonCollection {
    /**
     * The name of the collection. Use this
     * when creating, updating or retrieving values.
     */
    name: string;
    /**
     * The layout, which maps out how
     * to assemble values for storage and
     * retrieval so they are indexed
     * correctly.
     */
    layout: CollectionLayout;
    /**
     * Access patterns define how to retrieve
     * values on keys other than a document's
     * `_id` field
     */
    accessPatterns?: AccessPattern[];
    /**
     * The optional key path of an attribute
     * that will be copied to the TTL attribute
     * on collection documents.
     *
     * The value must be either:
     * * a Date object
     * * a string in ISO8601 format
     * * a number, in milliseconds since the Epoch
     *
     * (any other value will be ignored)
     *
     * The layout must also define `ttlAttribute` when you
     * specify `ttlKeyPath`, otherwise
     * an exception will be thrown when you create
     * the context.
     */
    ttlKeyPath?: KeyPath;
    /**
     * An optional function that is used to
     * generate IDs when the id property
     * is blank
     */
    idGenerator?: () => string;
}
/**
 * A root collection (or top-level collection) is
 * one without a parent. It is stored in a such a
 * way to enable the individual retrieval of items
 * or bulk retrieval of of child items by foreign
 * key.
 */
interface RootCollection extends CommonCollection {
    /**
     * A fixed value that distinguishes this as a
     * parent or 'root' collection (It is optional to
     * specify this)
     */
    type?: 'root';
}
/**
 * A child collection lets you store values related
 * to a parent collection (defined by [[RootCollection]])
 *
 * It's used to add data to an entity that is not always
 * needed for every access pattern, and can only be referenced
 * with regards to its parent.
 */
interface ChildCollection extends CommonCollection {
    /** A fixed value that must be defined to distinguish
     * this collection as a child collection
     */
    type: 'child';
    /**
     * The name of the parent collection. It must be added
     * to the same context as this child collection.
     */
    parentCollectionName: string;
    /**
     * The foreign key path
     */
    foreignKeyPath: KeyPath;
}
/**
 * A collection is a division based on entity type
 * for storage in a DynamoDB table. It tells dynaglue
 * how to store your documents and how to index them
 * for this entity type.
 *
 * Defining multiple
 * collections on the same DynamoDB table layout
 * allows you to store multiple data types in the
 * same DynamoDB table (i.e. a single table design).
 *
 * *Use the [[RootCollection]] or [[ChildCollection]] for specifying
 * a collection*
 */
declare type Collection = RootCollection | ChildCollection;

/**
 *
 * Extracted keys from access patterns. Used internally
 * to build stored values
 */
declare type ExtractKey = {
    type: 'partition' | 'sort' | 'ttl';
    key: string;
    valuePaths: KeyPath[];
    options: AccessPatternOptions;
};
/**
 *
 * Collection mapping (used internally)
 */
declare type CollectionDefinition = Collection & {
    wrapperExtractKeys: ExtractKey[];
};
/**
 * Root collection definition (used internally)
 */
declare type RootCollectionDefinition = RootCollection & {
    wrapperExtractKeys: ExtractKey[];
};
/**
 * Child collection definition (used internally)
 */
declare type ChildCollectionDefinition = ChildCollection & {
    wrapperExtractKeys: ExtractKey[];
};

/**
 * The internal representation of the context object.
 *
 * **This type should be considered opaque and subject to change.**
 */
interface DynaglueContext {
    ddb: DynamoDBClient;
    definitions: Map<string, CollectionDefinition>;
    rootDefinitions: Map<string, RootCollectionDefinition>;
    childDefinitions: Map<string, ChildCollectionDefinition>;
}

/**
 * An opaquely defined type
 */
declare type Opaque<K, T> = T & {
    __TYPE__: K;
};
/**
 * A dynaglue context. Use [[createContext]] to instantiate.
 *
 * The internal layout of this type may change over time.
 */
declare type Context = Opaque<'DynaglueContext', DynaglueContext>;
/**
 * Create a context object, with layouts and access patterns for
 * storing and retrieving data.
 *
 * @param dynamodb dynamodb instance, initialised with correct access key and region
 * @param collections a list of collection definitions to use with this context
 * @returns a context object
 *  @throws {ConfigurationException} when there is a configuration issue in the given collections
 */
declare function createContext(dynamodb: DynamoDBClient, collections: Collection[]): Context;

/**
 * A document stored in a collection. If an _id field
 * is not provided, one will be generated on insert.
 */
declare type DocumentWithId = {
    _id: string;
};
/**
 * A wrapped document, as it is written to DynamoDB.
 * You should not have to manipulate this type directly.
 */
declare type WrappedDocument<DocumentType extends DocumentWithId> = {
    type: string;
    value: DocumentType;
};
/**
 * A DynamoDB primary key
 */
declare type Key = {
    [key: string]: AttributeValue;
};
/**
 * The DynamoDB `Set` type
 */
declare type DynamoDBSet = Set<unknown>;

/**
 * Retrieve a top-level document by its `_id` field.
 *
 * @category Query
 *
 * @param context the context object
 * @param collectionName name of the collection to search
 * @param id the `_id` value of the root document
 * @param options the the set of options for the search
 * @param options.consistentRead search with strongly consistent reads
 * @returns the stored value, or `undefined` if not found
 * @throws {CollectionNotFoundException} when the collection is not found in the context
 */
declare function findById<DocumentType extends DocumentWithId>(context: Context, collectionName: string, id: string, options?: {
    consistentRead?: boolean;
}): Promise<DocumentType | undefined>;

/**
 * Insert a value into a collection. Adds an _id field to the value
 * if one is not provided using the `bson` ID generator (similar to
 * MongoDB IDs).
 *
 * @category Mutation
 *
 * @param context the context to use
 * @param collectionName name of the collection
 * @param value value to insert
 * @returns a copy of the inserted value, with appended `_id` field if not provided
 * @throws {ConflictException} when an item with the same _id already exists
 */
declare function insert<DocumentType extends DocumentWithId>(context: Context, collectionName: string, value: Record<string, unknown>): Promise<DocumentType>;

/**
 * The value for a condition expression
 */
declare type ConditionValue = unknown;
/**
 * A DynamoDB type, for the $type condition operator (see [[TypeCondition]])
 */
declare type DynamoDBType = 'S' | 'SS' | 'N' | 'NS' | 'B' | 'BS' | 'BOOL' | 'NULL' | 'L' | 'M';
/**
 * A filter or condition expression. This type
 * serves as the root of a object structure for specifying a
 * filter/condition expression which is translated to a DynamoDB
 * Conditional expression for use with some of the available
 * operations (in their underlying `ConditionExpression` or
 * `FilterExpression` fields.
 *
 * The syntax is very similar to Mongo's query syntax, but a little
 * bit stricter to help with type evaluation.
 *
 * A simple condition expression consists of key paths mapped to an operator.
 * For example, to test `profile.username = 'test@example.com'`, you would
 * construct an expression as follows:
 *
 * ```
 * {
 *   'profile.username': { $eq: 'test@example.com' },
 * }
 * ```
 *
 * This expression would be translated to DynamoDB's FilterExpression syntax as:
 *
 * `#attr0.username = :value0`
 *
 * (`profile` is a reserved word in DynamoDB and automatically mapped. `#attr0` and `:value0` will be
 * written to `ExpressionAttributeNames` and `ExpressionAttributeValues`
 * respectively for you).
 *
 * Multiple key paths in the same object are treated as an `AND` operation implicitly.
 *
 * * NOTE: For a full list of key path operators, see the [[KeyPathsAndClause]] type.
 *
 * You can also construct AND, OR and NOT expressions for more complex evaluations, e.g.:
 *
 * ```
 * {
 *   $or: [
 *   { $not: { 'location.value': { $lte: 0 } } },
 *   {
 *      $and: [
 *        { 'location.point.x': { $gt: 10 } },
 *        { 'location.point.y': { $gt: 1.3 } },
 *      ],
 *   }
 *   ],
 * };
 * ```
 * This would be written in DynamoDB's FilterExpression syntax as:
 *
 * `(NOT #attr0.#attr1 <= :value0) OR (#attr0.point.x > :value1 AND #attr0.point.y < :value2)`
 *
 * where:
 * * `#attr0` is mapped to 'location' (reserved word)
 * * `#attr1` is mapped to 'value' (reserved word)
 * * `:value0` is mapped to `0` (Expression Attribute Value)
 * * `:value1` is mapped to `10` (Expression Attribute Value)
 * * `:value2` is mapped to `1.3` (Expression Attribute Value)
 *
 */
declare type CompositeCondition = AndCondition | OrCondition | NotCondition | KeyPathsAndClause;
/**
 * An AND clause in a condition/filter expression. The `$and` array
 * must be the only key on the object, and contain at least one condition
 * clause.
 *
 * Each clause will be connected by an `AND` keyword.
 */
declare type AndCondition = {
    $and: CompositeCondition[];
};
/**
 * An OR clause in a condition/filter expression. The `$or` array
 * must be the only key on the object, and contain at least one condition
 * clause.
 *
 * Each clause will be connected by an `OR` keyword.
 */
declare type OrCondition = {
    $or: CompositeCondition[];
};
/**
 * A NOT clause in a condition/filter expression. The $not value must
 * be the only key on the object, and must contain a condition clause.
 */
declare type NotCondition = {
    $not: CompositeCondition;
};
/**
 * A clause that translates to the $eq operator. The key path this
 * appears on will be used in a `=` expression.
 *
 * e.g. `{ 'x.y': { $eq: 'once' } } => x.y = :value0 // (where :value0 is set to 'once')`
 */
declare type EqCondition = {
    $eq: ConditionValue;
};
/**
 * A clause that translates to the $eq operator. The key path this
 * appears on will be used in a `<>` expression.
 *
 * e.g. `{ 'x.y': { $neq: 5 } } => x.y <> :value0 // (where :value0 is set to 5)`
 */
declare type NotEqCondition = {
    $neq: ConditionValue;
};
/**
 * A clause that translates to the $gt operator. The key path this
 * appears on will be used in a `>` expression.
 *
 * e.g. `{ 'x.y': { $gt: 5 } } => x.y > :value0 // (where :value0 is set to 5)`
 */
declare type GtCondition = {
    $gt: ConditionValue;
};
/**
 * A clause that translates to the $gte operator. The key path this
 * appears on will be used in a `>=` expression.
 *
 * e.g. `{ 'x.y': { $gte: 5 } } => x.y >= :value0 // (where :value0 is set to 5)`
 */
declare type GtEqCondition = {
    $gte: ConditionValue;
};
/**
 * A clause that translates to the $lt operator. The key path this
 * appears on will be used in a `<` expression.
 *
 * e.g. `{ 'x.y': { $lt: 5 } } => x.y < :value0 // (where :value0 is set to 5)`
 */
declare type LtCondition = {
    $lt: ConditionValue;
};
/**
 * A clause that translates to the $lte operator. The key path this
 * appears on will be used in a `<=` expression.
 *
 * e.g. `{ 'x.y': { $lte: 5 } } => x.y <= :value0 // (where :value0 is set to 5)`
 */
declare type LtEqCondition = {
    $lte: ConditionValue;
};
/**
 * A clause that translates to the `$between` operator. The key path this
 * appears on will be used in a `BETWEEN ... AND ...` expression.
 */
declare type BetweenCondition = {
    $between: {
        $gte: ConditionValue;
        $lte: ConditionValue;
    };
};
/**
 * A clause that translates to the `$in` operator. The key path this
 * appears on will be used in a `IN (...)` expression.
 *
 * e.g. `{ x: { $in: ['a', 'b', 'c] } }` => `x IN (:v0, :v1, :v2)` (where `:v0 = 'a'`, `:v1 = 'b'`, etc)
 */
declare type InCondition = {
    $in: ConditionValue[];
};
/**
 * A clause that translates to the `$exists` operator. It's value
 * is a boolean expression:
 * * when `true`, it performs an `attribute_exists()` function on the associated property
 * * when `false`, it performs an `attribute_not_exists()` function on the associated property
 */
declare type ExistsCondition = {
    $exists: boolean;
};
/**
 * A clause that translates to the `$type` operator. It tests to see if the
 * associated attribute is of the specified type, returning a boolean result.
 *
 * It calls the `attribute_type()` function.
 *
 * The value is the DynamoDB type string to be tested (see https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/Expressions.OperatorsAndFunctions.html#Expressions.OperatorsAndFunctions.Functions)
 *
 * e.g. `{ x: { $type: 'L' } }` (see if `x` is a list) => `attribute_type(x, :v)` (where `:v = 'L'`)
 */
declare type TypeCondition = {
    $type: DynamoDBType;
};
/**
 * A clause that translates to the `$beginsWith` operator. It tests to see if a string
 * attribute begins with the specified string.
 *
 * It translates to the `begins_with()` function.
 *
 * e.g. `{ x: { $beginsWith: 'abc' } }` => `begins_with(x, :value0)` (where `:value0 = 'abc'`)
 */
declare type BeginsWithCondition = {
    $beginsWith: string;
};
/**
 * A clause that translates to the `$contains` operator. It tests to see if
 * the given value is contained in the attribute (which may be a Set or string).
 *
 * It translates to the `contains()` function.
 *
 * e.g. `{ x: { $contains: 'abc' } }` => `contains(x, :value0)` (where `:value0 = 'abc'`)
 */
declare type ContainsCondition = {
    $contains: string;
};
/**
 * A comparison condition. These clauses are the right-hand
 * side of a key-path in a condition expression.
 */
declare type ComparisonCondition = EqCondition | NotEqCondition | GtCondition | LtCondition | GtEqCondition | LtEqCondition | BetweenCondition | InCondition | ExistsCondition | TypeCondition | BeginsWithCondition | ContainsCondition;
/**
 * A key path condition clause. The keys in this object are key-paths
 * (`.` separated keys in the target object to be tested), while the values
 * are one of the operators listed in [[ComparisonCondition]].
 *
 * e.g. `{ 'profile.username': { $eq: 'user@example.com } }`
 * tests to see if the path `profile.username` in the target object
 * is equal to the string `'user@example.com'`.
 *
 * A key paths clause can contain multiple key paths with different operators;
 * the resulting DynamoDB clause is AND'ed together.
 *
 * e.g.
 * ```
 * {
 *   'profile.username': { $eq: 'user@example.com' },
 *   'suburbs': { $in: ['Placeville', 'Townsburb'] },
 *   'point.x': { $gt: 10.3 },
 * }
 * ```
 *
 * results in a DynamoDB condition expression like:
 * `profile.username = :value0 AND suburbs IN (:value1,:value2) AND point.x > :value3)`
 */
declare type KeyPathsAndClause = {
    [keyPath: string]: ComparisonCondition;
};

/**
 * The query operator to use in the `KeyConditionExpression` to `QueryItem`
 * as called by [[find]]
 */
declare type QueryOperator = 'match' | 'equals' | 'gte' | 'gt' | 'lte' | 'lt' | 'between';
/**
 * A find query. This is a map of key paths (specified dot-separated)
 * to values to match.
 *
 * The query you specify must match an access pattern on the collection,
 * otherwise an [[IndexNotFoundException]] will be thrown by [[find]]
 */
declare type FindQuery = {
    [matchKey: string]: string | [string, string];
};
/**
 * The results of a [[find]] operation.
 */
declare type FindResults<DocumentType extends DocumentWithId> = {
    /** The items in this batch of results */
    items: DocumentType[];
    /** The pagination token. This value is specified when
     * there are more results to fetch; pass it to another [[find]] call
     * to get the next batch. It will be `undefined` when there is no
     * more results.
     */
    nextToken?: Key;
};
/**
 * The options for a [[find]] operation
 */
declare type FindOptions = {
    /**
     * The sort key condition to use
     */
    queryOperator?: QueryOperator;
    limit?: number;
    /**
     * `true` (default) to scan the index forward, `false` to scan it backward
     */
    scanForward?: boolean;
    /**
     * An optional filter expression for the
     * find operation
     */
    filter?: CompositeCondition;
};
/**
 * Find an item using one of its collection's access patterns
 *
 * This method is very strict about how the query object is used. Failure
 * to specify it
 *
 * Each of its keys is one of the keys paths to a field in the collection objects
 * (separated by a full-stop `.`), with the values set to the value to
 * match against. The value is a direct string comparison, so only string
 * values can be looked up (the query object cannot contain nested values
 * or values of types other than string)
 *
 * Every path that is specified must be part of the same access pattern.
 * You cannot have extraneous fields to match or mix keys from different
 * access patterns.
 *
 * All of the key paths in the partition key part of the access pattern
 * must be specified.
 *
 * You do not have to specify all of the sort keys,
 * **but** because they are stored as a composite key, and we can only do
 * `begins_with` match, you must specify the keys in the order they appear,
 * only leaving off those at the end.
 *
 * For example, if your sort key is defined as:
 *
 * `[['state'], ['city'], ['street', 'name'], ['street', 'number']]`
 *
 * you could specify any of the following combinations:
 * - `{ state: '...' }`
 * - `{ state: '...', city: '...' }`
 * - `{ state: '...', city: '...', 'street.name': '...' }`
 * - `{ state: '...', city: '...', 'street.name': '...', 'street.number': '...' }`
 *
 * But these would be invalid:
 * - `{ city: '...' }`
 * - `{ state: '...', 'street.name': '...' }`
 * - `{ city: '...', 'street.name': '...', 'street.number': '...' }`
 *
 * @category Query
 *
 *
 * @param ctx the context object
 * @param collectionName the collection name
 * @param query an object, with each of its keys set to the path of
 * each field in the search object to match and their corresponding
 * values to match (see above)s
 * @param nextToken pagination token
 * @param options options controlling the query
 * @param options.queryOperator query operator to use (defaults to `begins_with()`)
 * @param options.limit number of records to return
 * @param options.scanForward scan direction, true for forward across the index (default) or false for backward
 * @param options.filter a filter expression
 * @returns an object containing the items found and a pagination token
 * (if there is more results)
 * @throws [[`CollectionNotFoundException`]] when the collection is not found in the context
 * @throws [[`InvalidQueryException`]] when the access pattern cannot be found for the specfied combination of query key paths
 */
declare function find<DocumentType extends DocumentWithId>(ctx: Context, collectionName: string, query: FindQuery, nextToken?: Key, options?: FindOptions): Promise<FindResults<DocumentType>>;

/**
 * Delete a root object using its `_id` field
 *
 * @category Mutation
 *
 * @param context the context object
 * @param collectionName the name of the collection
 * @param id the object to remove
 * @param options options to apply
 * @param options.condition an optional conditional expression that must be satifisfied for the update to proceed
 * @returns the deleted object (as stored in the database), or
 * `undefined` if not found
 * @throws {CollectionNotFoundException} when the collection is not found in the context
 */
declare function deleteById<DocumentType extends DocumentWithId>(context: Context, collectionName: string, id: string, options?: {
    condition?: CompositeCondition;
}): Promise<DocumentType | undefined>;

/**
 * Insert or replace a value in a collection.
 *
 * This operation differs from [[insert]] in that it does not check for the existence
 * of a document with the same `_id` value - it will replace whatever is there.
 *
 * @category Mutation
 *
 * @param context the context
 * @param collectionName the collection to update
 * @param value the document to insert or replace
 * @param options options to apply
 * @param options.condition an optional conditional expression that must be satifisfied for the update to proceed
 * @returns the inserted / replaced value
 * @throws {CollectionNotFoundException} when the collection is not found
 */
declare function replace<DocumentType extends DocumentWithId>(context: Context, collectionName: string, value: Record<string, unknown>, options?: {
    condition?: CompositeCondition;
}): Promise<DocumentType>;

/**
 * The results of a [[findChildren]] operation.
 */
declare type FindChildrenResults<DocumentType extends DocumentWithId> = {
    /** the items that were returned in this batch */
    items: DocumentType[];
    /** The pagination token. If this value is specified, it means
     * there is more results for the query. Provide it to another
     * call to `findChildren` to get the next set of results.
     */
    nextToken?: Key;
};
/**
 * The range expression to use on the sort key (`_id`) when
 * querying for child collection objects with
 * {@link findChildren}.
 */
declare type KeyRangeExpression = {
    op: 'gte' | 'gt' | 'lte' | 'lt' | 'begins_with';
    value: string;
} | {
    op: 'between';
    min: string;
    max: string;
};
/**
 * The options to a [[findChildren]] operation
 */
declare type FindChildrenOptions = {
    limit?: number;
    /**
     * `true` (default) to scan the index forward, `false` to scan it backward
     */
    scanForward?: boolean;
    /**
     * An optional filter expression for the
     * find operation
     */
    filter?: CompositeCondition;
    /**
     * The range of children to retrieve (on the child's ID value)
     */
    range?: KeyRangeExpression;
};
/**
 * Find all the child objects of a root (top-level) object.
 *
 * The parent collection is determined by the reference in the
 * child collection.
 *
 * This method has a `nextToken`, which is used for pagination - it
 * is returned when there is more values to retrieve. You should
 * write your code to repeatedly call findChildren with the nextToken
 * value of the previous call until it comes back undefined in order
 * to retrieve all the values.
 *
 * @category Query
 *
 * @param ctx the context
 * @param childCollectionName name of the child object collection
 * @param rootObjectId the `_id` of the root object
 * @param nextToken the next token from the previous call, or `undefined` if there are no more values
 * @param options the options to control the query
 * @param options.limit number of records to return
 * @param options.scanForward=true true for ascending index order
 * @param options.filter an optional filter expression to apply
 * @throws {CollectionNotFoundException} when the collection is not found in the context
 */
declare function findChildren<DocumentType extends DocumentWithId>(ctx: Context, childCollectionName: string, rootObjectId: string, nextToken?: Key, options?: FindChildrenOptions): Promise<FindChildrenResults<DocumentType>>;

/**
 * Retrieve a child item by its `_id` field and its parent `_id`.
 *
 * Child objects can only be directly retrieved with both the parent
 * and child _id. This limitation allows them to be stored without
 * an extra index, and enables the `findChildren` method to retrieve
 * all the children of a root object in the same call.
 *
 * @category Query
 *
 * @param context the context object
 * @param collectionName name of the collection to search
 * @param id the `_id` value
 * @param rootObjectId the _id of the root object
 * @param options the the set of options for the search
 * @param options.consistentRead search with strongly consistent reads
 * @returns the stored value, or `undefined` if not found
 * @throws {CollectionNotFoundException} when the collection is not found in the context
 */
declare function findChildById<DocumentType extends DocumentWithId>(context: Context, collectionName: string, id: string, rootObjectId: string, options?: {
    consistentRead?: boolean;
}): Promise<DocumentType | undefined>;

/**
 * Delete a child object using its `_id` field and its parents `_id`.
 *
 * @category Mutation
 *
 * @param context the context object
 * @param collectionName the name of the collection
 * @param id the child object to remove
 * @param rootObjectId the parent object id
 * @param options options to change behaviour of DynamoDB
 * @param options.condition a condition expression blocking the delete operation
 * @returns the deleted object (as stored in the database), or
 * `undefined` if not found
 * @throws {CollectionNotFoundException} when the collection is not found in the context
 */
declare function deleteChildById<DocumentType extends DocumentWithId>(context: Context, collectionName: string, id: string, rootObjectId: string, options?: {
    condition?: CompositeCondition;
}): Promise<DocumentType | undefined>;

/**
 * An update object, where the key paths are specified as keys and the values
 * as the new field values.
 *
 * Keep in mind, that although this is a partial update, you need to specify
 * all the keys for a composite key in an access pattern - it cannot partially
 * update composite index values.
 */
declare type SetValuesDocument = {
    [path: string]: unknown;
};
/**
 * A key path and value to set on a document as part
 * of an update operation. Specified as part of [[UpdateChangesDocument]],
 * this is a tuple in the form:
 *
 * `[key_path, new_value]`
 *
 * * *key_path* is the property path into the document to update (either
 *   as a dotted string path e.g. `'profile.name'` or a [[KeyPath]])
 * * *new_value* is the value to update the property path to (it cannot be
 *   undefined -- if you want to clear the value, see [[DeleteChange]])
 */
declare type SetChange = [string | KeyPath, unknown];
/**
 * A property path to delete on a document as part of a [[UpdateChangesDocument]],
 * specified either as a dotted string path (e.g. `'profile.name'`) or a [[KeyPath]]
 */
declare type RemoveChange = string | KeyPath;
/**
 * Add a number to a number property. Use negative values to subtract.
 */
declare type AddValueChange = [string | KeyPath, number];
/**
 * Append or delete the values from from to a set property.
 */
declare type AppendDeleteSetChange = [string | KeyPath, DynamoDBSet];
/**
 * A set of changes to perform in an [[updateById]] or [[updateChildById]]
 * operation as an object. Each property is optional, but at
 * least one change must be specified.
 */
declare type OperationUpdates = {
    /**
     * The list of key paths to set a value. This is a list
     * of tuples, in the form [key_path, new_value]. The
     * key paths can be specified in a string form e.g. `'profile.name'`
     * or array form e.g. `['profile', 'name']`
     */
    $set?: SetChange[];
    /**
     * The list of key paths to clear the value. Key paths
     * may be specified in string form e.g. `'profile.name'` or
     * array form e.g. `['profile', 'name']`
     */
    $remove?: RemoveChange[];
    /**
     * A set of key path + value pairs for which number
     * properties are to be added to.
     *
     * Key paths are specified in string (e.g. `profile.count`) or array (e.g. `['profile', 'count']) form.
     */
    $addValue?: AddValueChange[];
    /**
     * A set of key path + value pairs for which set
     * properties are to be appended.
     *
     * Key paths are specified in string (e.g. `profile.roles`) or array (e.g. `['profile', 'roles']) form.
     */
    $addToSet?: AppendDeleteSetChange[];
    /**
     * A set of key path + value pairs for which set
     * properties are to be appended.
     *
     * Key paths are specified in string (e.g. `profile.roles`) or array (e.g. `['profile', 'roles']) form.
     */
    $deleteFromSet?: AppendDeleteSetChange[];
};
declare type ChangesUpdates = {
    $setValues: SetValuesDocument;
};
/**
 * The set of updates to apply to a document. This can be
 * specified one of two ways:
 *
 * * an an object of key paths to values to set e.g. `{ 'profile.name': 'new name', 'status': 3 }`
 * * an operator object of changes to perform (see [[UpdateChangesDocument]])
 */
declare type Updates = OperationUpdates | ChangesUpdates;
/**
 * Update a document using its `_id`.
 *
 *
 * This operation allows you to do a partial update of a collection document i.e. without
 * specifying all the values (it uses DynamoDB`s `UpdateItem` operation).
 *
 * At this time, the `updates` value just updates specified key paths on the target document.
 *
 * If some of the update key paths are indexed values, the indexes will also be updated. Because
 * of this, you must specify all the key values in an access pattern to ensure indexes are
 * updated consistently.
 *
 * @param ctx context
 * @param collectionName the collection to update
 * @param objectId the _id value of the object to update
 * @param updates the set of updates to apply.
 * @param options options to apply (including conditional checks)
 *
 * @returns the updated object value in its entirety.
 *
 * @throws {@link CollectionNotFoundException} collection not found
 * @throws {@link InvalidUpdatesException} thrown when the updates object is invalid or incomplete
 * @throws {@link InvalidUpdateValueException} thrown when one of the update values is an invalid type
 *
 * @typeParam DocumentType your document type - used to set return type
 * @category Mutation
 *
 *
 */
declare function updateById<DocumentType extends DocumentWithId>(ctx: Context, collectionName: string, objectId: string, updates: Updates, options?: {
    condition?: CompositeCondition;
}): Promise<DocumentType>;

/**
 * Update a child document using its `_id` and parent `_id`.
 *
 * This operation allows you to do a partial update of a collection document i.e. without
 * specifying all the values (it uses DynamoDB`s `UpdateItem` operation).
 *
 * At this time, the `updates` value just updates specified key paths on the target document.
 *
 * If some of the update key paths are indexed values, the indexes will also be updated. Because
 * of this, you must specify all the key values in an access pattern to ensure indexes are
 * updated consistently.
 *
 * @category Mutation
 *
 * @param context the context
 * @param collectionName the collection to update
 * @param objectId the _id value of the object to update
 * @param parentObjectId: the _id value of the parent object
 * @param updates the set of updates to apply.
 * @param options options to apply
 * @param options.condition an optional conditional expression that must be satifisfied for the update to proceed
 * @returns the updated object value in its entirety.
 * @throws {@link CollectionNotFoundException} collection not found
 * @throws {@link InvalidUpdatesException} thrown when the updates object is invalid or incomplete
 * @throws {@link InvalidUpdateValueException} thrown when one of the update values is an invalid type
 */
declare function updateChildById<DocumentType extends DocumentWithId>(context: Context, collectionName: string, objectId: string, parentObjectId: string, updates: Updates, options?: {
    condition?: CompositeCondition;
}): Promise<DocumentType>;

declare type FindByIdWithChildrenResult<DocumentType extends DocumentWithId> = {
    /**
     * The root object.
     *
     * It may be blank if it is not in the current set of
     * results (this depends on how objects are organised
     * on your index, or if the root object exists at all).
     * If your result is paginated (i.e. there is a `nextToken`),
     * it is important to continue calling `findByIdWithChildren`
     * with the previous `nextToken` and keep checking `root`
     * on each set of results.
     */
    root?: DocumentType;
    /**
     * The child objects, organised by collection names.
     *
     * Similar to the root object, you must check each page
     * of results for children, which will come back based
     * on index order and scan direction.
     *
     * NOTE: These are typed as DocumentWithId: you will need
     * to typecast each of the child collections based on the name
     * you specify
     */
    children: {
        [collection: string]: DocumentWithId[];
    };
    /**
     * The pagination token. If this is not blank, you
     * need to call `findByIdWithChildren` with this
     * value to get the next page of results.
     */
    nextToken?: Key;
};
/**
 * Find a root object with one or more of its child collections.

 * It is used to implement the **adjacency list** access approach,
 * which is used to reduce round-trips by loading a parent
 * object and its associated child objects in the same call.
 *
 * (Because of this, it's important to understand adjacency lists
 * and why you would have modelled your collections in a
 * parent-child relationship to begin with in order to get
 * the desired performance outcomes)
 *
 * You can restrict the child collections to return by specifying
 * the `childCollectionNames` parameter. Leaving it blank retrieves
 * all the child objects.
 *
 * ## Performance
 *
 * Adjacency lists work by storing all the related items in the
 * same partition (key) i.e. with the same object ID as the
 * root object, and then keying the individual items in the
 * sort key.
 *
 * We can then use a `Query` call to return the root object
 * and child objects together. This means that the performance
 * and behaviour depends on however the items are stored in the index.
 *
 * For this reason it's vital to have
 * some understanding of how it is implemented, as well as
 * be aware of your most important access patterns, in order
 * to get the best performance (especially when you have
 * many child items or multiple child collections).
 *
 * ## How it works
 *
 * This method works by using a DynamoDB `Query` on the root
 * object ID (in the partition key). It restricts the child
 * objects returned by using a `BETWEEN` filter on the sort
 * key, bounded by lexigraphically lowest and highest collection
 * names.
 *
 * To ensure that only the specified collections are returned,
 * this is then filtered with a filter expression on the desired
 * child collection names. This is done with a `FilterExpression`
 * in the `Query` call.
 *
 * In the ideal scenario, all the
 * child objects are small enough be returned within DynamoDB's query limit.
 * In this case the `root` value will contain the specified parent object, and
 * the `children` value will be a key-value index on the child collection name
 * to its associated child objects.
 *
 * However, if the number of items exceeds the DynamoDB query limit (1MB
 * of total items scanned in the index), you will need to use
 * the `nextToken` to retrieve the next set of results. Because of
 * this, the `root` object may not be returned on the first call,
 * so you should check for it on each set of returned results.
 *
 * ## Working through an example
 *
 * Say you have a root object collection called *users*,
 * with two child collections, `addresses` and `credentials`.
 *
 * We have two main access patterns:
 * 1. Retrieve a user with all their addresses
 * 2. Retrieve a user will all their credentials
 *
 * ### Retrieving all the addresses
 *
 * For 1., where you request with child collection `addresses` only, a query like
 * this will be given to DynamoDB:
 *
 * ```
 * KeyConditionExpression: pk = "users|-|<userId>" AND sk BETWEEN "addresses|-|" AND 'users|-|'
 * FilterExpression: type IN ("users", "addresses")
 * ```
 *
 * Because of the way DynamoDB queries work, it will internally
 * scan over all the child objects (including the *credentials* objects).
 * The unwanted `credentials` objects will be filtered out, but
 * they will still count internally to the DynamoDB Query page limit
 * (about 1MB).
 *
 * The addresses objects will be returned first (the default ordering
 * goes in lexigraphic order), unless you specify `scanIndexForward: false`
 * to start at the end (where the *users* object is).
 *
 * If there isn't many unwanted *credentials*, the performance impact
 * will be neglible.
 *
 * If there any many *credentials* objects however, there could be many
 * blank pages returned until the root object is returned.
 *
 * ### Addressing the performance
 *
 * Firstly, we should change the scan behaviour to scan the index backwards.
 * This way, we'll get the root object in the first page. But DynamoDB
 * will still be scanning over the credentials objects internally.
 *
 * In cases like these, it may be better to break your request down into two
 * calls:
 *
 * * `findById` to return the root *users* object
 * * `findChildren` to return all the *addresses* object.
 *
 * However, this would negate the point of using a adjacency list
 * in the first place.
 *
 * If we look at our access patterns, we never want to retrieve
 * the credentials and addresses objects at the same time, so we
 * really want the *users* root object to appear between the two
 * sets of child objects.
 *
 * We could achieve this by changing the collection names so
 * that it is between the other two by adding a prefix to each collection name:
 *
 * ```
 * 0addresses
 * 1users
 * 2credentials
 * ```
 *
 * Then, we can fulfil (1) by requesting *1addresses* and scanning backwards
 * from *users*. Similarly, we can fulfill (2) by requesting *credentials*
 * and scanning in the default index order. In both cases, we won't be
 * scanning over unnecessary objects.
 *
 * @category Query
 *
 * @param ctx the dynaglue context object
 * @param rootCollectionName the root object collection
 * @param rootObjectId the root object ID
 * @param childCollectionNames the child collections to return - defaults to all child collections
 * @param nextToken the pagination token, to request the next page of results
 * @param options options to control the limit, scan order and extra filtering
 * @returns the root and/or child objects found, along with a `nextToken` if there is more results
 *
 */
declare const findByIdWithChildren: <DocumentType_1 extends DocumentWithId>(ctx: Context, rootCollectionName: string, rootObjectId: string, childCollectionNames?: string[], nextToken?: Key, options?: {
    limit?: number;
    scanForward?: boolean;
}) => Promise<FindByIdWithChildrenResult<DocumentType_1>>;

/**
 * The collection and ID of a root or child
 * item to retrieve with #transactFindByIds
 */
declare type BatchFindByIdDescriptor = {
    /** The collection containing the item */
    collection: string;
    id: string;
    parentId?: string;
};
/**
 * The response to a [[batchFindByIds]] request. The
 * retrieved documents are stored in a map, organised by
 * collection name.
 *
 * Any unprocessed request keys are included in the
 * unprocessedDescriptors list. You will need to submit
 * another request to obtain these.
 */
declare type BatchFindByIdsResponse = {
    documentsByCollection: {
        [collection: string]: DocumentWithId[];
    };
    /**
     * Unprocessed keys - you will need to submit
     * another request for these subsequently
     */
    unprocessedDescriptors: BatchFindByIdDescriptor[];
};
/**
 * Find multiple items by their primary key
 * in bulk. May specify items over multiple
 * tables and collections.
 *
 * @category Batch Operations
 */
declare const batchFindByIds: (ctx: Context, items: BatchFindByIdDescriptor[], options?: {
    consistentReadTableNames?: string[];
}) => Promise<BatchFindByIdsResponse>;

/**
 * A replace (PutItem) request to perform with
 * [[batchReplaceDelete]].
 *
 * Performs the equivalent operation as [[replace]],
 * which is a full insert or update on the primary
 * key of the document.
 */
declare type BatchReplaceDescriptor = {
    /** The root or child collection */
    collection: string;
    /** The operation to perform (always `'replace'`) */
    op: 'replace';
    /** The document of the item to be replaced */
    replaceItem: DocumentWithId;
};
/**
 * A replace (DeleteItem) request to perform with
 * [[batchReplaceDelete]].
 *
 * Performs the equivalent operation as [[deleteById]]/[[deleteChildById][,
 * which is a full delete on the primary key of a document
 */
declare type BatchDeleteDescriptor = {
    /** The root or child collection */
    collection: string;
    /** The operation to perform */
    op: 'delete';
    /** The item identifier */
    id: string;
    /**
     * For child items, this is the parent object identifier
     * (mandatory for root
     * items, left blank for child items).
     */
    parentId?: string;
};
/**
 * A replace (PutItem) or delete (DeleteItem) request
 * to perform with [[batchReplaceDelete]]
 */
declare type BatchReplaceDeleteDescriptor = BatchReplaceDescriptor | BatchDeleteDescriptor;
/**
 * Response from {@link batchReplaceDelete}. Contains
 * unprocessed items.
 */
declare type BatchReplaceDeleteResponse = {
    /**
     * The items that were not processed in this call. You will
     * need to resubmit them to the API in a subsequent call.
     */
    unprocessedDescriptors: BatchReplaceDeleteDescriptor[];
};
/**
 * Replace multiple items in bulk (effectively a bulk
 * `replaceItem()` call).
 *
 * You can update multiple items across different
 * tables and collections in the same call.
 *
 * @category Batch Operations
 */
declare const batchReplaceDelete: (ctx: Context, descriptors: BatchReplaceDeleteDescriptor[]) => Promise<BatchReplaceDeleteResponse>;

/**
 * An element in the expression parse tree, used
 * to assist error messaging.
 */
declare type ParseElement = {
    type: 'array';
    index: number;
} | {
    type: 'object';
    key: string;
};

/**
 * Thrown when insert() is called with a specified _id
 * that already exists
 */
declare class ConflictException extends VError {
    constructor(message: string, id: string);
}
/**
 * Thrown when an `_id` value is specified that is not
 * valid (_id values must be a string).
 */
declare class InvalidIdException extends VError {
    constructor(id: unknown);
}
/**
 * Thrown when the child object to be inserted has a missing or invalid
 * parent key value.
 */
declare class InvalidParentIdException extends VError {
    constructor(parentId: unknown, collectionName: string, parentCollectionName: string);
}
/**
 * Thrown when the collection specified cannot be found in the context, or
 * isn't of the expected type (root or child) for the called API.
 */
declare class CollectionNotFoundException extends VError {
    constructor(collection: string);
}
/**
 * Thrown when the index is not found for the provided query.
 */
declare class IndexNotFoundException extends VError {
    constructor(index: string);
}
/**
 * Thrown during context construction to indicate a configuration
 * issue in the specified collections, access patterns or layouts.
 */
declare class ConfigurationException extends VError {
    constructor(message: string, options?: Options);
}
/**
 * Thrown on invalid field values that are provided during
 * persistence (insert/update/replace).
 */
declare class InvalidIndexedFieldValueException extends VError {
    constructor(message: string, { collection, keyPath }: {
        collection: string;
        keyPath: string[];
    });
}
/**
 * Thrown on an invalidly specified query provided to a
 * `find()` operation.
 */
declare class InvalidQueryException extends VError {
    constructor(message: string, { collection, query }: {
        collection: string;
        query: Record<string, unknown>;
    });
}
/**
 * Thrown when the updates provided to an `update()` operation
 * are invalid.
 */
declare class InvalidUpdatesException extends VError {
    constructor(message: string);
}
/**
 * Thrown when the update values provided in an `update()` operation
 * are invalid (e.g. `undefined`)
 */
declare class InvalidUpdateValueException extends VError {
    constructor(path: string, message: string);
}
/**
 * Thrown when there is a problem with the expression given
 * as a `FilterExpression` or `ConditionExpression`.
 */
declare class InvalidCompositeConditionException extends VError {
    constructor(message: string, parsePath: ParseElement[]);
}
/**
 * An exception thrown when a bad set of
 * [[BatchReplaceDeleteDescriptor]] is given to
 * [[batchReplaceDelete]]
 */
declare class InvalidBatchReplaceDeleteDescriptorException extends VError {
    constructor(message: string, info?: Record<string, unknown>);
}

export { AccessPattern, AccessPatternOptions, AddValueChange, AndCondition, AppendDeleteSetChange, BatchDeleteDescriptor, BatchFindByIdDescriptor, BatchFindByIdsResponse, BatchReplaceDeleteDescriptor, BatchReplaceDeleteResponse, BatchReplaceDescriptor, BeginsWithCondition, BetweenCondition, ChangesUpdates, ChildCollection, Collection, CollectionLayout, CollectionNotFoundException, ComparisonCondition, CompositeCondition, ConditionValue, ConfigurationException, ConflictException, ContainsCondition, Context, DocumentWithId, DynamoDBSet, DynamoDBType, EqCondition, ExistsCondition, FindByIdWithChildrenResult, FindChildrenOptions, FindChildrenResults, FindOptions, FindQuery, FindResults, GtCondition, GtEqCondition, InCondition, IndexNotFoundException, InvalidBatchReplaceDeleteDescriptorException, InvalidCompositeConditionException, InvalidIdException, InvalidIndexedFieldValueException, InvalidParentIdException, InvalidQueryException, InvalidUpdateValueException, InvalidUpdatesException, Key, KeyPath, KeyPathsAndClause, KeyRangeExpression, LtCondition, LtEqCondition, NormaliserFunction, NotCondition, NotEqCondition, OperationUpdates, OrCondition, ParseElement, PrimaryIndexLayout, QueryOperator, RemoveChange, RootCollection, SecondaryIndexLayout, SetChange, SetValuesDocument, TypeCondition, Updates, WrappedDocument, batchFindByIds, batchReplaceDelete, createContext, deleteById, deleteChildById, find, findById, findByIdWithChildren, findChildById, findChildren, insert, replace, updateById, updateChildById };
